import io
import uuid
import requests
from concurrent.futures import ThreadPoolExecutor
from typing import List, Callable, Optional, Iterable

import google.cloud.storage.bucket

from gs_chunked_io.config import default_chunk_size, gs_max_parts_per_compose, writer_retries
from gs_chunked_io.async_collections import AsyncSet


class Writer(io.IOBase):
    """
    Writable stream on top of GS blob.

    Chunks of `chunk_size` bytes are uploaded as individual blobs and composed into a multipart object using the API
    described here: https://cloud.google.com/storage/docs/composite-objects. An attempt is made to clean up
    incomplete or aborted writes.

    if `part_callback` is provided, will be called for each part with part number, part blob key, and part data as
    arguments.

    Chunks are uploaded with concurrency equal to `threads`. If the maximum number of uploads is currently in progress,
    `write` blocks until an upload slot becomes available.

    Concurrency can be disabled by passing in`threads=None`.
    """
    def __init__(self,
                 key: str,
                 bucket: google.cloud.storage.bucket.Bucket,
                 chunk_size: int=default_chunk_size,
                 part_callback: Optional[Callable[[int, str, bytes], None]]=None,
                 threads: Optional[int]=4):
        self.key = key
        self.bucket = bucket
        self.chunk_size = chunk_size
        self._part_callback = part_callback
        self._part_names: List[str] = list()
        self._buffer = bytearray()
        self._current_part_number = 0
        self._closed = False
        self._upload_id = uuid.uuid4()
        if threads is not None:
            assert 1 <= threads
            max_workers = max(threads, 8)
            self.executor: Optional[ThreadPoolExecutor] = ThreadPoolExecutor(max_workers=max_workers)
            self.future_chunk_uploads = AsyncSet(self.executor, concurrency=threads)
        else:
            self.executor = None
        try:
            bucket.blob
        except AttributeError:
            raise TypeError("Expected instance of google.cloud.storage.bucket.Bucket, or similar.")

    @property
    def closed(self) -> bool:
        return self._closed

    def _put_part(self, part_number: int, data: bytes):
        if data:
            part_name = self._name_for_part_number(part_number)
            for tries_remaining in range(writer_retries - 1, -1, -1):
                try:
                    self.bucket.blob(part_name).upload_from_file(io.BytesIO(data))
                    break
                except requests.exceptions.ConnectionError:
                    if 0 == tries_remaining:
                        raise
            self._part_names.append(part_name)
            if self._part_callback:
                self._part_callback(part_number, part_name, data)
        else:
            raise ValueError(f"data for part_number={part_number} must not be empty!")

    def writable(self) -> bool:
        return True

    def write(self, data: bytes):
        self._buffer += data
        while len(self._buffer) >= self.chunk_size:
            if self.executor:
                for _ in self.future_chunk_uploads.consume_finished():
                    pass
                self.future_chunk_uploads.put(self._put_part, self._current_part_number, self._buffer[:self.chunk_size])
            else:
                self._put_part(self._current_part_number, self._buffer[:self.chunk_size])
            del self._buffer[:self.chunk_size]
            self._current_part_number += 1

    def writelines(self, *args, **kwargs):
        raise NotImplementedError()

    def close(self):
        if not self._closed:
            self._closed = True
            if self._buffer:
                self._put_part(self._current_part_number, self._buffer)
            self._compose_dest_blob()
            if self.executor:
                self.executor.shutdown()

    def _compose_dest_blob(self):
        self.wait()
        part_names = self._sorted_part_names(self._part_names)
        part_numbers = [len(part_names)]
        parts_to_delete = set(part_names)
        while gs_max_parts_per_compose < len(part_names):
            name_groups = [names for names in _iter_groups(part_names, group_size=gs_max_parts_per_compose)]
            new_part_numbers = list(range(part_numbers[-1], part_numbers[-1] + len(name_groups)))
            if self.executor:
                part_names = self.executor.map(self._compose_parts,
                                               name_groups,
                                               [self._name_for_part_number(n) for n in new_part_numbers])
                part_names = self._sorted_part_names(part_names)
            else:
                part_names = [self._compose_parts(names, self._name_for_part_number(new_part_number))
                              for names, new_part_number in zip(name_groups, new_part_numbers)]
            parts_to_delete.update(part_names)
            part_numbers = new_part_numbers
        self._compose_parts(part_names, self.key)
        self._delete_parts(parts_to_delete)

    def _delete_parts(self, part_names):
        def _del(name):
            self.bucket.blob(name).delete()

        if self.executor:
            for _ in self.executor.map(_del, part_names):
                pass
        else:
            for name in part_names:
                _del(name)

    def _compose_parts(self, part_names, dst_part_name) -> str:
        blobs = [self.bucket.blob(name) for name in part_names]
        self.bucket.blob(dst_part_name).compose(blobs)
        return dst_part_name

    def _name_for_part_number(self, part_number) -> str:
        """
        Compose a Google Storage object name corresponding to `part_number`.
        Since Google Storage shards based on object name, it is more optimal to use names that being with uniformly
        distributed random strings, such as a uuid.
        See: https://cloud.google.com/blog/products/gcp/optimizing-your-cloud-storage-performance-google-cloud-performance-atlas  # noqa
        """
        part_id = uuid.uuid4()
        return f"{part_id}.{self._upload_id}.gs-chunked-io-part.%06i" % part_number

    def _sorted_part_names(self, part_names: Iterable[str]) -> List[str]:
        """
        Sort names by part number.
        """
        return [p[1]
                for p in sorted([(name.rsplit(".", 1)[1], name) for name in part_names])]

    def wait(self):
        """
        Wait for all concurrent uploads to finish.
        If `executor` is None, this method does nothing.
        """
        if self.executor:
            for _ in self.future_chunk_uploads.consume():
                pass

    def abort(self):
        """
        Close, cancel write operations, and delete written parts from `self.bucket`.
        """
        if not self._closed:
            self._closed = True
            if self.executor:
                self.future_chunk_uploads.abort()
            self._delete_parts(self._part_names)

    def __del__(self, *args, **kwargs):
        self.abort()
        super().__del__(*args, **kwargs)

    def seek(self, *args, **kwargs):
        raise OSError()

    def tell(self, *args, **kwargs):
        raise NotImplementedError()

    def read(self, *args, **kwargs):
        raise OSError()

    def truncate(self, *args, **kwargs):
        raise NotImplementedError()

def _iter_groups(lst: list, group_size=32):
    for i in range(0, len(lst), group_size):
        yield lst[i:i + group_size]

class AsyncPartUploader:
    """
    Concurrently put parts in any order. `put_part` blocks when concurrent uploads equals or exceeds
    `concurrent_uploads`. The executor should have equal or more threads than `concurrent_uploads`.
    """

    def __init__(self,
                 key: str,
                 bucket: google.cloud.storage.bucket.Bucket,
                 threads: int=4):
        assert 1 <= threads
        self._writer = Writer(key, bucket, threads=threads)
        self.executor = ThreadPoolExecutor(max_workers=threads)
        self.future_chunk_uploads = AsyncSet(self.executor, concurrency=threads)

    def put_part(self, part_number: int, data: bytes):
        for _ in self.future_chunk_uploads.consume_finished():
            pass
        self.future_chunk_uploads.put(self._writer._put_part, part_number, data)

    def close(self):
        for _ in self.future_chunk_uploads.consume():
            pass
        self._writer.close()
        self.executor.shutdown()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()



"""
Provide objects to manage results from concurrent operations using ThreadPoolExecutor.
"""
import multiprocessing
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from typing import Any, List, Set, Callable, Generator, Optional


class _AsyncCollection:
    def __init__(self, executor: ThreadPoolExecutor, concurrency: int=0):
        self.executor = executor
        self.concurrency = concurrency
        assert 0 < self.concurrency

    def __len__(self):
        raise NotImplementedError()

    def __bool__(self) -> bool:
        return bool(len(self))

    def put(self, func: Callable, *args, **kwargs):
        raise NotImplementedError()

    def get(self) -> Any:
        raise NotImplementedError()

    def consume(self) -> Generator[Any, None, None]:
        raise NotImplementedError()

    def consume_finished(self) -> Generator[Any, None, None]:
        raise NotImplementedError()

    def _running(self) -> Set[Future]:
        futures = getattr(self, "_futures", list())
        return {f for f in futures if not f.done()}

    def abort(self):
        futures = getattr(self, "_futures", list())
        for f in futures:
            f.cancel()
        for _ in as_completed(futures):
            pass

    def __del__(self):
        self.abort()

class AsyncSet(_AsyncCollection):
    """
    Unordered collection providing results of concurrent operations. Up to `concurrency` operations are executed in
    parallel.
    """
    def __init__(self, executor: ThreadPoolExecutor, concurrency: int=multiprocessing.cpu_count()):
        super().__init__(executor, concurrency)
        self._futures: Set[Future] = set()

    def __len__(self):
        return len(self._futures)

    def put(self, func: Callable, *args, **kwargs):
        """
        Submit a new operation to the executor. Block until the number of in-progress operations becomes less than or
        equal to `concurrency`.
        """
        running = len(self._running())
        if running >= self.concurrency:
            self._wait(running - self.concurrency)
        f = self.executor.submit(func, *args, **kwargs)
        self._futures.add(f)

    def get(self) -> Any:
        """
        Remove and return any result. Wait until a result is available.
        """
        for r in self.consume():
            return r

    def consume(self) -> Any:
        """
        Remove and yield all results as available, waiting as needed.
        """
        for f in as_completed(self._futures):
            self._futures.remove(f)
            yield f.result()

    def consume_finished(self) -> Generator[Any, None, None]:
        """
        Remove and yield results from completed operations.
        """
        for f in self._futures.copy():
            if f.done():
                self._futures.remove(f)
                yield f.result()

    def _wait(self, count: int=-1):
        for i, f in enumerate(as_completed(self._running())):
            f.result()
            if i == count:
                break

class AsyncQueue(_AsyncCollection):
    """
    FIFO queue providing results of concurrent operations in the order they were submitted. Up to `concurrency`
    operations are executed in parallel. New operations are executed concurrently as available results are consumed.
    """
    def __init__(self, executor: ThreadPoolExecutor, concurrency: int=multiprocessing.cpu_count()):
        super().__init__(executor, concurrency)
        self._futures: List[Future] = list()  # type: ignore
        self._to_be_submitted: List[Any] = list()

    def __len__(self):
        return len(self._to_be_submitted) + len(self._futures)

    def _submit_futures(self):
        while len(self._futures) < self.concurrency:
            try:
                func, args, kwargs = self._to_be_submitted.pop(0)
            except IndexError:
                return
            f = self.executor.submit(func, *args, **kwargs)
            self._futures.append(f)

    def put(self, func: Callable, *args, **kwargs):
        """
        Schedule a new operation for execution.
        """
        self._to_be_submitted.append((func, args, kwargs))
        self._submit_futures()

    def get(self) -> Any:
        """
        Remove and return the next result in order. Wait until available.
        """
        res = None
        for f in as_completed(self._futures[:1]):
            res = self._futures.pop(0).result()
        self._submit_futures()
        return res

    def consume(self) -> Generator[Any, None, None]:
        """
        Remove and yield all results in order, waiting as needed.
        """
        while self._to_be_submitted or self._futures:
            yield self.get()



import io
from math import ceil
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

from google.cloud.storage.blob import Blob

from gs_chunked_io.config import default_chunk_size, reader_retries
from gs_chunked_io.async_collections import AsyncQueue


_BLOB_CHUNK_SIZE_UNIT = 262144

class Reader(io.IOBase):
    """
    Readable stream on top of GS blob. Bytes are fetched in chunks of `chunk_size`.
    Chunks are downloaded with concurrency equal to `threads`.

    Concurrency can be disabled by passing in`threads=None`.
    """
    def __init__(self, blob: Blob, chunk_size: int=default_chunk_size, threads: Optional[int]=2):
        assert chunk_size >= 1
        if blob.size is None:
            blob.reload()
        if blob.chunk_size is None:
            # Induce google.cloud.storage.blob to use either google.resumable_media.requests.ChunkedDownload or
            # google.resumable_media.requests.RawChunkedDownload, which do not attempt to perform data-integrity checks
            # for chunk downloads (checksum headers are not available for chunks).
            blob.chunk_size = ceil(chunk_size / _BLOB_CHUNK_SIZE_UNIT) * _BLOB_CHUNK_SIZE_UNIT
        self.blob = blob
        self.chunk_size = chunk_size
        self._buffer = bytearray()
        self._pos = 0
        self.number_of_chunks = ceil(self.blob.size / self.chunk_size)
        self._unfetched_chunks = [i for i in range(self.number_of_chunks)]
        if threads is not None:
            assert 1 <= threads
            self.executor: Optional[ThreadPoolExecutor] = ThreadPoolExecutor(max_workers=threads)
            self.future_chunk_downloads = AsyncQueue(self.executor, threads)
            for chunk_number in self._unfetched_chunks:
                self.future_chunk_downloads.put(self._fetch_chunk, chunk_number)
        else:
            self.executor = None

    def _fetch_chunk(self, chunk_number: int) -> bytes:
        start_chunk = chunk_number * self.chunk_size
        end_chunk = start_chunk + self.chunk_size - 1
        if chunk_number == (self.number_of_chunks - 1):
            expected_part_size = self.blob.size % self.chunk_size
        else:
            expected_part_size = self.chunk_size
        for _ in range(reader_retries):
            data = self.blob.download_as_string(start=start_chunk, end=end_chunk)
            if expected_part_size == len(data):
                return data
        raise ValueError("Unexpected part size")

    def readable(self):
        return True

    def read(self, size: int=-1) -> bytes:
        if -1 == size:
            size = self.blob.size
        if size + self._pos > len(self._buffer):
            del self._buffer[:self._pos]
            if self.executor:
                while size > len(self._buffer) and len(self.future_chunk_downloads):
                    self._buffer += self.future_chunk_downloads.get()
            else:
                while size > len(self._buffer) and len(self._unfetched_chunks):
                    self._buffer += self._fetch_chunk(self._unfetched_chunks.pop(0))
            self._pos = 0

        ret_data = bytes(memoryview(self._buffer)[self._pos:self._pos + size])
        self._pos += len(ret_data)
        return ret_data

    def readinto(self, buff: bytearray) -> int:
        d = self.read(len(buff))
        bytes_read = len(d)
        buff[:bytes_read] = d
        return bytes_read

    def seek(self, *args, **kwargs):
        raise OSError()

    def tell(self, *args, **kwargs):
        raise NotImplementedError()

    def truncate(self, *args, **kwargs):
        raise NotImplementedError()

    def write(self, *args, **kwargs):
        raise OSError()

    def close(self):
        if self.executor:
            self.executor.shutdown()
        super().close()

def for_each_chunk(blob: Blob, chunk_size: int=default_chunk_size, threads: Optional[int]=2):
    """
    Fetch chunks and yield in order. Chunks are downloaded with concurrency equal to `threads`
    """
    reader = Reader(blob, chunk_size=chunk_size)
    if threads is not None:
        assert 1 <= threads
        with ThreadPoolExecutor(max_workers=threads) as e:
            future_chunk_downloads = AsyncQueue(e, concurrency=threads)
            for chunk_number in reader._unfetched_chunks:
                future_chunk_downloads.put(reader._fetch_chunk, chunk_number)
            for chunk in future_chunk_downloads.consume():
                yield chunk
    else:
        for chunk_number in reader._unfetched_chunks:
            yield reader._fetch_chunk(chunk_number)

def for_each_chunk_async(blob: Blob, chunk_size: int=default_chunk_size, threads: int=4):
    """
    Fetch chunks with concurrency equal to `threads`, yielding results as soon as available.
    Results may be returned in any order.
    """
    assert 1 <= threads
    reader = Reader(blob, chunk_size)

    def fetch_chunk(chunk_number):
        data = reader._fetch_chunk(chunk_number)
        return chunk_number, data

    with ThreadPoolExecutor(max_workers=threads) as e:
        for chunk_number, chunk in e.map(fetch_chunk, range(reader.number_of_chunks)):
            yield chunk_number, chunk



import os
import glob
import subprocess
import warnings
from setuptools import setup, find_packages

install_requires = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "requirements.txt"))]

with open("README.md") as fh:
    long_description = fh.read()

def get_version():
    filepath = os.path.join(os.path.dirname(__file__), "gs_chunked_io", "version.py")
    if os.path.isfile(filepath):
        # In source distributions or builds, version is available in the generated gs_chunked_io/version.py file
        with open(filepath) as fh:
            version = dict()
            exec(fh.read().strip(), version)
            return version['__version__']
    else:
        p = subprocess.run(["git", "describe", "--tags", "--match", "v*.*.*"], stdout=subprocess.PIPE)
        if 128 == p.returncode:
            warnings.warn('There are no git tags with version information. '
                          'To tag the first commit as v0.0.0 use '
                          '`git tag --annotate "v0.0.0" $(git rev-list --max-parents=0 HEAD) -m "v0.0.0"`')
            return "0"
        else:
            p.check_returncode()
            out = p.stdout.decode("ascii").strip()
            if "-" in out:
                out = out.split("-", 1)[0]
            assert out.startswith("v")
            return out[1:]

setup(
    name='gs-chunked-io',
    version=get_version(),
    description='Streaming read/writes to Google Storage blobs with ascynchronous buffering.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/xbrianh/gs-chunked-io.git',
    author='Brian Hannafious',
    author_email='bhannafi@ucsc.edu',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    scripts=glob.glob('scripts/*'),
    zip_safe=False,
    install_requires=install_requires,
    platforms=['MacOS X', 'Posix'],
    test_suite='test',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7'
    ]
)
